/**********************************************************************/
/**********************************************************************/
/* The GIM project                                                    */
/**********************************************************************/
/* FILENAME:         gct_data.cpp                                     */
/* ORIGINATION DATE: 11/17/04                                         */
/* ORIGINATOR:       D Smith                                          */
/* MODULE:           none                                             */
/* LICENSE:          gpl (www.gnu.org/copyleft/gpl.html)              */
/**********************************************************************/
/* GIM Copyright (C)2004, 2005                                        */
/*   Darrin Smith - darrin[at]dbcservices[dot]com                     */
/*   Mark Roth - GIM[dot]mroth[at]gmail[dot]com                       */
/*   And others.                                                      */
/* GIDForums is a registered trademark                               */
/*                                                                    */
/* Find GIM at (put savannah site here)                               */
/* Discuss GIM at GIDForums[dot]com                                   */
/* Contact GIM at GIMmailer[at]gmail[dot]com                          */
/*                                                                    */
/**********************************************************************/
/*               This file is part of GIM.                            */
/* gct_data is free software. You can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published  */
/* by the Free Software Foundation, either version 2 of the License,  */
/* or (at your option) any later version.                             */
/* gct_data is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY' without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      */
/* GNU General Public License for more details.                       */
/*                                                                    */
/* You should have received a copy of the GNU General Public License  */
/* along with gct_data. If not, write to :                            */
/*       the Free Software Foundation, Inc.                           */
/*       59 Temple Place                                              */
/*       Suite 330                                                    */
/*       Boston, MA  02111-1307  USA                                  */
/*                                                                    */
/*  Licence in distribution, COPYING.txt                              */
/*                                                                    */
/**********************************************************************/
/* DESCRIPTION:                                                       */
/* This file contains the interface functions needed for GIMcontacts  */
/* to interact with the GIM database.  You may only use the functions */
/* marked as public to conform to the GIM interfacing rules.          */
/* Any other useage may render your program inoperable by updates to  */
/* gim.h.                                                             */
/**********************************************************************/
/**********************************************************************/
#include "gim.h"
#include "dlist.h"

#include <string.h>
#include <ctype.h>
#include <stdio.h>

//Global Variables
dlist*	ct_list = 0;				//This is the main list for the data	
dlist*	ct_flist = 0;				//This is the filter list when a filter is applied

//MUST be defined according to gim.h
int 	MAX_NAME;
int 	MAX_PHONE;
int 	MAX_WEB;
int 	MAX_ADDRESS;
int 	MAX_COUNTRY;
int		MAX_ZIP;
int		MAX_NOTES;

//This is a flag to check to see if a filter is active or not.
bool	FILTER_SET;

//This holds the sort structure.
contact*	sorter;


/*--------------------------------------------------*/
/* Utility functions (private functions)            */
/* These functions are helper functions and should  */
/* not be called outside of this module             */
/* -------------------------------------------------*/

/*******************************************************************************/
/* compare is necessary to use the doubly linked list as defined by dlist.h.   */
/* This is a rather lengthy implementation due to the custom sort feature.     */
/*******************************************************************************/
int compare(void* data1, void* data2)
{
	contact*	entry1 = (contact*) data1;
	contact*	entry2 = (contact*) data2;
	int			current = 1;
	int			order;
	char*		string1;
	char*		string2;
	int			rv;
	bool		done = FALSE;
	
	do{
		if(sorter->type == current){
			if(rv = entry1->type - entry2->type)
				return rv;
		}
		else{
			if(*(sorter->lname) == current){
				string1 = entry1->lname;
				string2 = entry2->lname;
				order = *(sorter->lname+1);
			}
			else if(*(sorter->fname) == current){
				string1 = entry1->fname;
				string2 = entry2->fname;
				order = *(sorter->fname+1);
			}
			else if(*(sorter->mname) == current){
				string1 = entry1->mname;
				string2 = entry2->mname;
				order = *(sorter->mname+1);
			}
			else if(*(sorter->cname) == current){
				string1 = entry1->cname;
				string2 = entry2->cname;
				order = *(sorter->cname+1);
			}
			else if(*(sorter->city) == current){
				string1 = entry1->city;
				string2 = entry2->city;
				order = *(sorter->city+1);
			}
			else if(*(sorter->state) == current){
				string1 = entry1->state;
				string2 = entry2->state;
				order = *(sorter->state+1);
			}

			else if(*(sorter->country) == current){
				string1 = entry1->country;
				string2 = entry2->country;
				order = *(sorter->country+1);
			}
			else
				done = TRUE;
			if( (rv = strcmp(string1,string2) ) )
				if(order == SORT_ASCEND)
					return rv;
				else
					return -rv;
		}
		current++;
	}while(!done);
	
	return 0;
}
	

/*******************************************************************************/
/* range is another necessary function needed to use the doubly linked list as */
/* defined by dlist.h.  Again, this is rather lengthy due to the custom sort   */
/* feature.                                                                    */
/*******************************************************************************/
float range(void* begin, void* end, void* eval)
{
	contact*	entry1 = (contact*) begin;
	contact*	entry2 = (contact*)	end;
	contact*	range = (contact*) eval;
	char*		string1;
	char*		string2;
	char*		string3;
	
	float	value1;
	float	value2;
	if(sorter->type == 1){
		value1 = range->type - entry2->type;
		value2 = entry2->type - entry1->type;
	}
	else{
		if(*(sorter->lname) == 1){
			string1 = entry1->lname;
			string2 = entry2->lname;
			string3 = range->lname;
		}
		else if(*(sorter->fname) == 1){
			string1 = entry1->fname;
			string2 = entry2->fname;
			string3 = range->fname;
		}
		else if(*(sorter->mname) == 1){
			string1 = entry1->mname;
			string2 = entry2->mname;
			string3 = range->mname;
		}
		else if(*(sorter->cname) == 1){
			string1 = entry1->cname;
			string2 = entry2->cname;
			string3 = range->cname;
		}
		else if(*(sorter->city) == 1){
			string1 = entry1->city;
			string2 = entry2->city;
			string3 = range->city;
		}
		else if(*(sorter->state) == 1){
			string1 = entry1->state;
			string2 = entry2->state;
			string3 = range->state;
		}
		else if(*(sorter->country) == 1){
			string1 = entry1->country;
			string2 = entry2->country;
			string3 = range->country;
		}
		value1 = *string3 - *string2;
		value2 = *string2 - *string1;
	}
	if(value2 == 0)
		return 0.0;
	else
		return (value1/value2);
}	


/*******************************************************************************/
/* copy_malloc is a small utility that allocates a string that is exactly big  */
/* enough to hold the passed string and copies the passed string into it.      */
/*******************************************************************************/
char* copy_malloc(char* source){

	char*	new_string;

	if( ! (new_string = (char*) malloc(strlen(source)+1) ) )
		new_string = 0;
	else
		strcpy(new_string,source);

	return new_string;
}


/*******************************************************************************/
/* read_entry is a helper function to loading a file.  The files use a NULL    */
/* character to split entries and a newline to split records.  This will       */
/* hopefully limit file corruption up to the newline.                          */
/*******************************************************************************/
int read_entry(FILE* fp, char* string, int max, int STRING_FLAG=1)
{
	char* 	parse = string;
	char	read;
	int		counter = 0;
	
	do{
		read = (char) fgetc(fp);
		if(read == EOF)
			return EOF;
		if(isprint(read) || ( (!STRING_FLAG) && (read))){
			*parse++ = read;
			counter++;
		}
	}while((read != 0) && (read != '\n') && (counter<max) );
	*parse = 0;
	
	return read;
}


/*******************************************************************************/
/* resort is called when a new sort structure is passed.  Due to the nature of */
/* the dlist, it is simpler to rebuild a new list and then set ct_list to point*/
/* to it.  This obviously means that sorts can be slow on large data sets      */
/*******************************************************************************/
void resort()
{
	dlist*	new_list = new dlist;
	dlist*	temp;
	void*	location;
	
	if(ct_list->home()){
		do{
			location = ct_list->get_data();
			new_list->add_sort(location,compare,range);
		}while(ct_list->next());
	}
	temp = ct_list;
	ct_list = new_list;
	delete temp;
	ct_list->home();
}


/*******************************************************************************/
/* ct_clear simply places empty strings at all entries in the passed contact   */
/* structure.  This is called by ct_malloc                                     */
/*******************************************************************************/
void ct_clear(contact* new_contact)
{
	*(new_contact->lname) = 0;
	*(new_contact->fname) = 0;
	*(new_contact->mname) = 0;
	*(new_contact->cname) = 0;
	
	*(new_contact->home) = 0;
	*(new_contact->work) = 0;
	*(new_contact->mobil) = 0;
	*(new_contact->fax) = 0;
	*(new_contact->other1) = 0;
	*(new_contact->other2) = 0;
	
	*(new_contact->email) = 0;
	*(new_contact->url) = 0;
	
	*(new_contact->address1) = 0;
	*(new_contact->address2) = 0;
	*(new_contact->city) = 0;
	*(new_contact->state) = 0;
	*(new_contact->zipcode) = 0;
	*(new_contact->country) = 0;
	
	*(new_contact->notes) = 0;
	new_contact->type = PERSONAL;
	
}


/*******************************************************************************/
/* ct_find_address is used when a ct_delete is called and a filter is active.  */
/* The delete needs to happen on the main list, not the filter list.  This     */
/* function places the main list at the proper location.                       */
/*******************************************************************************/
int ct_find_address(contact* entry)
{
	ct_home();
	do{
		if(entry == (contact*) ct_list->get_data())
			return 1;
	}while(ct_next());
	return 0;
}


/*******************************************************************************/
/* match is used by ct_filter.  It returns 1 if the filter criteria matches the*/
/* string and 0 if not.  This was done quite quickly and could have some holes.*/
/* I am almost positive that a filter of *s would return a false for a string  */
/* simms, because it will hang on the first s.                                 */
/*******************************************************************************/
int match(char* filter, char* string)
{
	char*	fparse = filter;
	char*	sparse = string;
	
	if(!*filter)
		return 1;
	while(*fparse){
		if((*fparse == '?') || (*sparse == *fparse)){
			fparse++;
			sparse++;
		}
		else if(*fparse == '*'){
			fparse++;
			while(*sparse != *fparse){
				sparse++;
				if(!*sparse)
					return 0;
			}
		}
		else
			return 0;
	}
	if(!*sparse) 
		return 1;
	else
		return 0;
}

/* END - Utility Functions (Private Functions)      */
















/*--------------------------------------------------*/
/* Contact functions (public functions)             */
/* These functions are protyped in gim.h and it is  */
/* important that they are implemented as defined   */
/* -------------------------------------------------*/
int ct_new(	int max_name, int max_phone, int max_web, int max_address, int max_country, int max_zip, int max_notes)
{
	ct_delete();						//Free old database if it exists
	
	ct_list = new dlist;				//Initialize double linked storage list
	
	//Set all maximums for internal & external usage.
	MAX_NAME = max_name;
	MAX_PHONE = max_phone;
	MAX_WEB = max_web;
	MAX_ADDRESS = max_address;
	MAX_COUNTRY = max_country;
	MAX_ZIP = max_zip;
	MAX_NOTES = max_notes;
	
	FILTER_SET = FALSE;					//Filter is not set currently.
	sorter = ct_sort_malloc();			//Allocate memory for sort definition
	ct_sort_default(sorter);			//Apply defualt sort
	
	return GE_OKAY;
}


int ct_delete()
{
	contact*	temp_entry;
	
	
	if(ct_list){						//Check to make sure list is defined
		if(FILTER_SET)					//Clear filter if set
			ct_filter_clear();
			
		//Parse through each item and remove allocated memory
		if(ct_list->home()){
			do{
				temp_entry = (contact*)ct_list->get_data();
				ct_free(temp_entry);
			}while(ct_list->next());
		}
		delete ct_list;					//Unallocate doubly linked list memory
		ct_sort_free(sorter);			//Unallocate sort definition memory 
		ct_list = 0;
		ct_flist = 0;
		return GE_OKAY;
	}
	else 
		return GE_FREE;					
}


contact* ct_malloc()
{
	contact* new_contact = (contact*) malloc(sizeof(contact));		//Allocate memory for structure

	//Allocate maximum memory for each item in the new structure
	new_contact->lname = (char*) malloc(MAX_NAME);	
	new_contact->fname = (char*) malloc(MAX_NAME);
	new_contact->mname = (char*) malloc(MAX_NAME);
	new_contact->cname = (char*) malloc(MAX_NAME);
	
	new_contact->home = (char*) malloc(MAX_PHONE);
	new_contact->work = (char*) malloc(MAX_PHONE);
	new_contact->mobil = (char*) malloc(MAX_PHONE);
	new_contact->fax = (char*) malloc(MAX_PHONE);
	new_contact->other1 = (char*) malloc(MAX_PHONE);
	new_contact->other2 = (char*) malloc(MAX_PHONE);
	
	new_contact->email = (char*) malloc(MAX_WEB);
	new_contact->url = (char*) malloc(MAX_WEB);
	
	new_contact->address1 = (char*) malloc(MAX_ADDRESS);
	new_contact->address2 = (char*) malloc(MAX_ADDRESS);
	new_contact->city = (char*) malloc(MAX_COUNTRY);
	new_contact->state = (char*) malloc(MAX_COUNTRY);
	new_contact->zipcode = (char*) malloc(MAX_ZIP);
	new_contact->country = (char*) malloc(MAX_COUNTRY);
	
	new_contact->notes = (char*) malloc(MAX_NOTES);
	
	//Initilize all values to be safe!
	ct_clear(new_contact);
	
	return(new_contact);
}


int ct_free(contact* entry)
{

	//Free all memory from each entry
	free(entry->lname);
	free(entry->fname);
	free(entry->mname);
	free(entry->cname);
	
	free(entry->home);
	free(entry->work);
	free(entry->mobil);
	free(entry->fax);
	free(entry->other1);
	free(entry->other2);
	
	free(entry->email);
	free(entry->url);
	
	free(entry->address1);
	free(entry->address2);
	free(entry->city);
	free(entry->state);
	free(entry->zipcode);
	free(entry->country);
	
	free(entry->notes);
	
	//Finally free structure memory
	free(entry);
}


int ct_put(contact* entry, sort_type s)
{
	contact*	new_entry = (contact*) malloc(sizeof(contact));	//Container for new entry
	ct_filter_clear();						//Clear any set filters
	
	//Copy name entries
	new_entry->lname = copy_malloc(entry->lname);
	new_entry->fname = copy_malloc(entry->fname);
	new_entry->mname = copy_malloc(entry->mname);
	new_entry->cname = copy_malloc(entry->cname);
	
	//Copy Phone number entries
	new_entry->home = copy_malloc(entry->home);
	new_entry->work = copy_malloc(entry->work);
	new_entry->mobil = copy_malloc(entry->mobil);
	new_entry->fax = copy_malloc(entry->fax);
	new_entry->other1 = copy_malloc(entry->other1);
	new_entry->other2 = copy_malloc(entry->other2);
	
	//Copy Mail & Web entries
	new_entry->email = copy_malloc(entry->email);
	new_entry->url = copy_malloc(entry->url);
	
	//Copy Address entries
	new_entry->address1 = copy_malloc(entry->address1);
	new_entry->address2 = copy_malloc(entry->address2);
	new_entry->city = copy_malloc(entry->city);
	new_entry->state = copy_malloc(entry->state);
	new_entry->zipcode = copy_malloc(entry->zipcode);
	new_entry->country = copy_malloc(entry->country);
	
	//Miscellaneous entries
	new_entry->notes = copy_malloc(entry->notes);
	new_entry->type = entry->type;
	
	//Insert into the list based on insertion flag
	switch(s){
		case LAST:				//Insert as the last item
			ct_list->add_data((void*)new_entry);
			break;
		case FIRST:				//Insert as the first item
			ct_list->home();
			ct_list->add_atpos((void*)new_entry);
			break;
		default:				//Do a full sort insert
			ct_list->add_sort((void*)new_entry,compare,range);
			break;
	}
	return(GE_OKAY);
}	


int ct_get(contact* entry)
{
	contact* data_entry;
	
	//If the filter is set get the data from that list otherwise use the full list.
	if(FILTER_SET)
		data_entry = (contact*) ct_flist->get_data();
	else
		data_entry = (contact*) ct_list->get_data();
	
	//Copy name entries
	strcpy(entry->lname,data_entry->lname);
	strcpy(entry->fname,data_entry->fname);
	strcpy(entry->mname,data_entry->mname);
	strcpy(entry->cname,data_entry->cname);
	
	//Copy Phone number entries
	strcpy(entry->home,data_entry->home);
	strcpy(entry->work,data_entry->work);
	strcpy(entry->mobil,data_entry->mobil);
	strcpy(entry->fax,data_entry->fax);
	strcpy(entry->other1,data_entry->other1);
	strcpy(entry->other2,data_entry->other2);
	
	//Copy Mail & Web entries
	strcpy(entry->email,data_entry->email);
	strcpy(entry->url,data_entry->url);
	
	//Copy Address entries
	strcpy(entry->address1,data_entry->address1);
	strcpy(entry->address2,data_entry->address2);
	strcpy(entry->city,data_entry->city);
	strcpy(entry->state,data_entry->state);
	strcpy(entry->zipcode,data_entry->zipcode);
	strcpy(entry->country,data_entry->country);
	
	//Miscellaneous entries
	strcpy(entry->notes,data_entry->notes);
	entry->type = data_entry->type;
	
	return(GE_OKAY);
}


int ct_mod(contact* entry)
{
	//This is a pretty simple function because despite its name it does not modify the current
	//record at all.  It deletes the current record and inserts a new record.
	//This may seem like overkill, but this is actually easier than modifying the entry and
	//doing a resort on the entire data.
	
	//Remove the current record.
	ct_remove();
	
	//Add the new record
	ct_put(entry);
}


int ct_remove()
{

	if(FILTER_SET){
		//If the filter is set this becomes a bit more complicated.
		//We need to set the main list pointer to the data to be deleted.
		ct_find_address((contact*)ct_flist->get_data());
		ct_filter_clear();				//Clear filter so we are working on the main list.
	}
	if(!ct_is_empty()){
		ct_free((contact*)ct_list->get_data());
		ct_list->rem_data();
	}
}



int ct_load(char* name)
{
	FILE*		fp;
	contact*	entry;
	char*		string = (char*) malloc(MAX_NOTES);
	int			rv = 0;
	char		version[20];
	
	if( !(fp = fopen(name,"r")) )
		return GE_FILENAME;
	entry = ct_malloc();
	
	/*** Check Header String ***/
	read_entry(fp,string,100);
	if(strcmp(string,"GIM Contact Data File"))
		return GE_INVALID;
	
	/*** Check Version ***/
	sprintf(version,"%d.%d.%d%c",GID_RELEASE,GID_FEATURE,GID_CHANGE,GID_VERSION);
	read_entry(fp,string,20);
	if(strcmp(string,version))
		return GE_VERSION;
		
	/*** Load Maximum size settings ***/
	rv = read_entry(fp,string,10);
	MAX_NAME = atoi(string);
	if(!rv){
		rv = read_entry(fp,string,10);
		MAX_PHONE = atoi(string);
	}
	if(!rv){
		rv = read_entry(fp,string,10);
		MAX_WEB = atoi(string);
	}
	if(!rv){
		rv = read_entry(fp,string,10);
		MAX_ADDRESS = atoi(string);
	}
	if(!rv){
		rv = read_entry(fp,string,10);
		MAX_COUNTRY = atoi(string);
	}
	if(!rv){
		rv = read_entry(fp,string,10);
		MAX_ZIP = atoi(string);
	}
	if(!rv){
		rv = read_entry(fp,string,10);	
		MAX_NOTES = atoi(string);
	}
		
	/*** Load Sort Structure ***/
	rv = read_entry(fp,sorter->lname,3,0);
	if(!rv)
		rv = read_entry(fp,sorter->fname,3,0);
	if(!rv)
		rv = read_entry(fp,sorter->mname,3,0);
	if(!rv)
		rv = read_entry(fp,sorter->cname,3,0);
	if(!rv)
		rv = read_entry(fp,sorter->city,3,0);
	if(!rv)
		rv = read_entry(fp,sorter->state,3,0);
	if(!rv)
		rv = read_entry(fp,sorter->country,3,0);
	if(!rv){
		rv = read_entry(fp,string,2,0);
		sorter->type = (ct_type)atoi(string);
	}
	
	/*** Load Data - A newline will automatically jump to the next record ***/
	while(rv != EOF){
		rv = read_entry(fp,string,MAX_NAME);
		if(rv == EOF)
			break;
		strcpy(entry->lname,string);
		if(!rv){
			rv = read_entry(fp,string,MAX_NAME);
			strcpy(entry->fname,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_NAME);
			strcpy(entry->mname,string);
		}

		if(!rv){
			rv = read_entry(fp,string,MAX_NAME);
			strcpy(entry->cname,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_PHONE);
			strcpy(entry->home,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_PHONE);
			strcpy(entry->work,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_PHONE);
			strcpy(entry->mobil,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_PHONE);
			strcpy(entry->fax,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_PHONE);
			strcpy(entry->other1,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_PHONE);
			strcpy(entry->other2,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_WEB);
			strcpy(entry->email,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_WEB);
			strcpy(entry->url,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_ADDRESS);
			strcpy(entry->address1,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_ADDRESS);
			strcpy(entry->address2,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_COUNTRY);
			strcpy(entry->city,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_COUNTRY);
			strcpy(entry->state,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_ZIP);
			strcpy(entry->zipcode,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_COUNTRY);
			strcpy(entry->country,string);
		}
		if(!rv){
			rv = read_entry(fp,string,MAX_NOTES);
			strcpy(entry->notes,string);
		}
		if(!rv){
			rv = read_entry(fp,string,2);
			entry->type = (ct_type)atoi(string);
		}
		ct_put(entry,LAST);
	}
	ct_free(entry);
	fclose(fp);
	free(string);
	FILTER_SET = FALSE;
	
	return GE_OKAY;
	
}


int ct_save(char* name)
{	
	FILE*		fp;
	contact*	entry;
	
	if( !(fp = fopen(name,"w") ) )
		return GE_FILENAME;
	if(ct_home()){
		
		/*** Write out Header information ***/
		fputs("GIM Contact Data File\n",fp);
		fprintf(fp,"%d.%d.%d%c\n",GID_RELEASE,GID_FEATURE,GID_CHANGE,GID_VERSION);
		
		/*** Write out Maximum string size globals ***/
		fprintf(fp,"%d",MAX_NAME);fputc(0,fp);
		fprintf(fp,"%d",MAX_PHONE);fputc(0,fp);
		fprintf(fp,"%d",MAX_WEB);fputc(0,fp);
		fprintf(fp,"%d",MAX_ADDRESS);fputc(0,fp);
		fprintf(fp,"%d",MAX_COUNTRY);fputc(0,fp);
		fprintf(fp,"%d",MAX_ZIP);fputc(0,fp);
		fprintf(fp,"%d",MAX_NOTES);fputc('\n',fp);
		
		/*** Write out Sort Structure ***/
		fputs(sorter->lname,fp);fputc(0,fp);
		fputs(sorter->fname,fp);fputc(0,fp);
		fputs(sorter->mname,fp);fputc(0,fp);
		fputs(sorter->cname,fp);fputc(0,fp);
		fputs(sorter->city,fp);fputc(0,fp);
		fputs(sorter->state,fp);fputc(0,fp);
		fputs(sorter->country,fp);fputc(0,fp);
		fputc(sorter->type+48,fp);fputc('\n',fp);
		
		/*** Write out each record - Field seperated by a NULL, Records seperated by a newline ***/
		do{
			entry = (contact*) ct_list->get_data();
			fputs(entry->lname,fp);fputc(0,fp);
			fputs(entry->fname,fp);fputc(0,fp);
			fputs(entry->mname,fp);fputc(0,fp);
			fputs(entry->cname,fp);fputc(0,fp);
			fputs(entry->home,fp);fputc(0,fp);
			fputs(entry->work,fp);fputc(0,fp);
			fputs(entry->mobil,fp);fputc(0,fp);
			fputs(entry->fax,fp);fputc(0,fp);
			fputs(entry->other1,fp);fputc(0,fp);
			fputs(entry->other2,fp);fputc(0,fp);
			fputs(entry->email,fp);fputc(0,fp);
			fputs(entry->url,fp);fputc(0,fp);
			fputs(entry->address1,fp);fputc(0,fp);
			fputs(entry->address2,fp);fputc(0,fp);
			fputs(entry->city,fp);fputc(0,fp);
			fputs(entry->state,fp);fputc(0,fp);
			fputs(entry->zipcode,fp);fputc(0,fp);
			fputs(entry->country,fp);fputc(0,fp);
			fputs(entry->notes,fp);fputc(0,fp);
			fputc(entry->type+48,fp);
			fputc('\n',fp);
		}while(ct_next());
	}
	fclose(fp);
	return GE_OKAY;	
}

int ct_home()
{
	if(FILTER_SET)
		return(ct_flist->home());
	else
		return(ct_list->home());
}

int ct_end()
{
	if(FILTER_SET)
		return(ct_flist->end());
	else
		return(ct_list->end());
}

int ct_next(int amount)
{
	int rv;
		do{
		if(FILTER_SET)
			rv = ct_flist->next();
		else
			rv = ct_list->next();
	}while( (--amount) && rv );
	return rv;
}
int ct_prev(int amount)
{
	int rv;
	
	do{
		if(FILTER_SET)
			rv = ct_flist->prev();
		else
			rv = ct_list->prev();
	}while( (--amount) && rv );
	
	return rv;
}

void* ct_mark()
{
	if(FILTER_SET)
		return ct_flist->set_mark();
	else
		return(ct_list->set_mark());
}

int ct_goto(void* location)
{
	if(FILTER_SET)
		ct_flist->goto_mark(location);
	else
		ct_list->goto_mark(location);
	return GE_OKAY;
}

int ct_is_empty()
{
	if(FILTER_SET)
		if(ct_flist->is_empty())
			return GE_EMPTY;
		else
			return GE_OKAY;
	else
		if(ct_list->is_empty())
			return GE_EMPTY;
		else
			return GE_OKAY;
}


int ct_filter(contact* filter)
{
	contact*	current;
	bool		FOUND = FALSE;
	
	ct_filter_clear();				//Clear any existing filter rules
	ct_flist = new dlist;
	if(ct_home()){		
		do{
			current = (contact*) ct_list->get_data();
			if(match(filter->lname,current->lname))
				if(match(filter->fname,current->fname))
					if(match(filter->cname,current->cname))
						if(match(filter->home,current->home))
							if(match(filter->work,current->work))
								if(match(filter->mobil,current->mobil))
									if(match(filter->fax,current->fax))
										if(match(filter->email,current->email))
											if(match(filter->url,current->url))
												if(match(filter->address1,current->address1))
													if(match(filter->address2,current->address2))
														if(match(filter->city,current->city))
															if(match(filter->state,current->state))
																if(match(filter->country,current->country))
																	if((filter->type == ALL) || (filter->type == current->type)){
																		ct_flist->add_data((void*)current);
																		FOUND = TRUE;
																	}
		}while(ct_next());
		if(FOUND)
			FILTER_SET = TRUE;
	}
	return GE_OKAY;
}


int ct_filter_clear()
{
	if(FILTER_SET){
		FILTER_SET = FALSE;
		delete ct_flist;
	}
	return GE_OKAY;
}


contact* ct_sort_malloc()
{
	contact* new_contact = (contact*) malloc(sizeof(contact));		//Allocate Memory for structure
	
	
	//Allocate Memory for sort string for applicable entries  (only need 3 bytes)
	new_contact->lname = (char*) malloc(3);
	new_contact->fname = (char*) malloc(3);
	new_contact->mname = (char*) malloc(3);
	new_contact->cname = (char*) malloc(3);
	new_contact->city = (char*) malloc(3);
	new_contact->state = (char*) malloc(3);
	new_contact->country = (char*) malloc(3);
	
	//Set all values to 0 (no sort defined)
	*(new_contact->lname) = 0;
	*(new_contact->fname) = 0;
	*(new_contact->mname) = 0;
	*(new_contact->cname) = 0;
	*(new_contact->city) = 0;
	*(new_contact->state) = 0;
	*(new_contact->country) = 0;
	new_contact->type = ALL;
	
	return(new_contact);	
}


int ct_sort_free(contact* entry)
{
	//Free all memory used in the applicable entries.
	free(entry->lname);
	free(entry->fname);
	free(entry->mname);
	free(entry->cname);
	free(entry->city);
	free(entry->state);
	free(entry->country);
	
	//Free the structure itself
	free(entry);
	
	return GE_OKAY;
}


int ct_sort(contact* new_sort)
{
	//Copy the sort passed to the private sort structure
	strcpy(sorter->lname,new_sort->lname);
	strcpy(sorter->fname,new_sort->fname);
	strcpy(sorter->mname,new_sort->mname);
	strcpy(sorter->cname,new_sort->cname);
	strcpy(sorter->city,new_sort->state);
	strcpy(sorter->state,new_sort->country);
	strcpy(sorter->country,new_sort->country);
	sorter->type = new_sort->type;
	
	//Resort data based upon new sort.
	resort();
	
	return GE_OKAY;
}



int ct_sort_get(contact* temp_sort)
{	
	strcpy(temp_sort->lname,sorter->lname);
	strcpy(temp_sort->fname,sorter->fname);
	strcpy(temp_sort->mname,sorter->mname);
	strcpy(temp_sort->cname,sorter->cname);
	strcpy(temp_sort->city,sorter->city);
	strcpy(temp_sort->state,sorter->state);
	strcpy(temp_sort->country,sorter->country);
	temp_sort->type = sorter->type;
	
	return GE_OKAY;
}


int ct_sort_default(contact* new_sort)
{
	//Default sort is last name, first name & middle name - all ascending
	
	//Set the order for the default sort
	*(new_sort->lname) = 1;
	*(new_sort->fname) = 2;
	*(new_sort->mname) = 3;
	
	//Set all sort types as ascending
	*(new_sort->lname+1) = SORT_ASCEND;
	*(new_sort->fname+1) = SORT_ASCEND;
	*(new_sort->mname+1) = SORT_ASCEND;
	
	//Null terminate strings
	*(new_sort->lname+2) = 0;
	*(new_sort->fname+2) = 0;
	*(new_sort->mname+2) = 0;
	
	return GE_OKAY;
}











