/**********************************************************************/
/**********************************************************************/
/* The GIM project                                                    */
/**********************************************************************/
/* FILENAME:         dlist.h                                          */
/* ORIGINATION DATE: 5/21/04                                          */
/* ORIGINATOR:       Darrin Smith                                     */
/* MODULE:           none                                             */
/* LICENSE:          gpl (www.gnu.org/copyleft/gpl.html)              */
/**********************************************************************/
/* GIM Copyright (C)2004, 2005                                        */
/*   Darrin Smith - darrin[at]dbcservices[dot]com                     */
/*   Mark Roth - GIM[dot]mroth[at]gmail[dot]com                       */
/*   And others.                                                      */
/* GIDForums is a registered trademark                               */
/*                                                                    */
/* Find GIM at (put savannah site here)                               */
/* Discuss GIM at GIDForums[dot]com                                   */
/* Contact GIM at GIMmailer[at]gmail[dot]com                          */
/*                                                                    */
/***********************************************************************
/*               This file is part of GIM.                            */
/* dlist.h is free software. You can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published  */
/* by the Free Software Foundation, either version 2 of the License,  */
/* or (at your option) any later version.                             */
/* dlist.h is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY' without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      */
/* GNU General Public License for more details.                       */
/*                                                                    */
/* You should have received a copy of the GNU General Public License  */
/* along with dlist.h. If not, write to :                             */
/*       the Free Software Foundation, Inc.                           */
/*       59 Temple Place                                              */
/*       Suite 330                                                    */
/*       Boston, MA  02111-1307  USA                                  */
/*                                                                    */
/*  Licence in distribution, COPYING.txt                              */
/*                                                                    */
/***********************************************************************
/*** DESCRIPTION:                                                    ***
/*** This is a double linked list implementation which makes use of  ***
/*** a "hash" function to determine the closest starting point for a ***
/*** linear search.  This initial guess reduces the time needed to   ***
/*** search/insert the list.                                         ***
/***********************************************************************
/*** NOTES                                                           ***
/*** Since this list works on void* (unknown data type), two helper  ***
/*** Functions are needed.  These are compare and range.  These are  ***
/*** included below based upon an integer value being needed.  You   ***
/*** can adapt these and include them in your code as necessary.     ***
/***                                                                 ***
/*** int compare(void* data1, void* data2)                           ***
/*** {                                                               ***
/*** 	if(*( (int*) data1 ) > *( (int*) data2) )                    ***
/*** 		return 1;                                                ***
/*** 	if(*( (int*) data1 ) < *( (int*) data2) )                    ***
/*** 		return -1;                                               ***
/*** 	return 0;                                                    ***
/*** }                                                               ***
/***                                                                 ***
/***                                                                 ***
/*** float range(void* begin, void* end, void* eval)                 ***
/*** {                                                               ***
/*** 	float	value1;                                              ***
/*** 	float	value2;                                              ***
/***	                                                             ***
/*** 	value1 = *( (int*) eval ) - *( (int*) begin);                ***
/*** 	value2 = *( (int*) end ) - *( (int*) begin);                 ***
/***                                                                 ***
/*** 	if(value2 == 0)                                              ***
/*** 		return 0.0;                                              ***
/*** 	else                                                         ***
/*** 		return (value1/value2);	                                 ***
/*** }                                                               ***
/**********************************************************************/
/**********************************************************************/

#ifndef __DLIST_H
#define	__DLIST_H

#include <stdlib.h>
#define	UNIQUE	1
#define	TRUE	1
#define	FALSE	0

class dlist
{
private:
	struct data{
		void* 	info;
		data*	next;
		data*	prev;
	};
	data*	head;			//Always points to first element in list
	data*	tail;			//Always points to last element in list
	data*	current;		//Always points to current position in list
	long	size;			//Holds the true(based on 1) size of list
	long	index;			//This is the current index
	int 	CLOSE;			//Set initially to determine hash_search range
	bool	is_unique;		//Boolean that tells whether this is a unique only entries
public:

	//********************************
	//*** CONSTRUCTOR & DESTRUCTOR ***
	//********************************
	dlist(int type = 0, int search_range = 200){
 		head = tail = current = 0;
 		size = 0;
		index = -1;
		CLOSE = search_range;
		if(type)
			is_unique = TRUE;
		else
			is_unique = FALSE;
		  
  	};
  	
	~dlist(){
 		home();
 		while(!is_empty())
			rem_data();
	};
	
	
	//**************************
	//*** MOVEMENT FUNCTIONS ***
	//**************************
	int	next(){
 		if(!is_empty()){
			if(current->next){
				current = current->next;
				index++;
				return 1;
			}
		}
		return 0;
		
	};
	
	int	prev(){
		if(!is_empty())
			if(current->prev){
				current = current->prev;
				index--;
				return 1;		
			}
		return 0;
			
 	};
 	
	int	home(){
 		if(is_empty())
			return 0;
		current = head;
		index = 0;
		return 1;
	};
	
	int	end(){
 		if(is_empty())
			return 0;
		current = tail;
		index = size-1;
		return 1;
	};
	
	void goto_index(long where){
		if(where == index)
			return;
		if( where > index )
			while( where>index && next() );
		else
			while( where<index && prev() );
	}
	
	
	//**********************
	//*** DATA FUNCTIONS ***
	//**********************
	void add_data(void* new_data){
		
		data* temp;
				
		temp = new data;
		temp->next = 0;
		temp->prev = 0;
		temp->info = new_data;
		if(end()){
			current->next = temp;
			temp->prev = current;
		}
        else{
			head = temp;
		}
		size++;
		index = size-1;
		current = temp;
		tail = current;
	};
	
	void add_atpos(void* new_data){
		
		if(is_empty())
			add_data(new_data);
		else{	
			data* temp = new data;
			temp->info = new_data;
		
			if(is_home()){
				temp->next = current;
				current->prev = temp;
				temp->prev = 0;
				head = temp;
				home();
			}
			else{
				current->prev->next = temp;
				temp->prev = current->prev;
				temp->next = current;
				current->prev = temp;
			}
			current = temp;
			size++;
		}
	};
	
	int add_sort(void* new_data, int (*compare)(void*,void*), float (*range)(void*,void*,void*)){
		int search;
		
		search = find_data(new_data, (compare), (range) );
		
		if(search){
			if( (!is_unique) || (search == -1) ) 
				add_atpos(new_data);
			else
				return -1;	//Identical record with unique flag - return -1, don't overwrite data.
		}
		else
			add_data(new_data);
		return 0;
	};
	
	void* get_data(){
		if( (!is_empty()) && (current) )
			return(current->info);
   		else
			return 0;
	};
	
	void* get_atpos(long where){
		if(!is_empty()){
			goto_index(where);
			return get_data();
		}
		return 0;
	};
	
	void rem_data(){
	
		data* temp;
		
		if(!is_empty()){
			if(is_home()){
				head = head->next;
				if(head)
					head->prev = 0;
				delete current;
				home();
			}
			else{
				temp = current->prev;
				temp->next = current->next;
				if(!is_end())
					current->next->prev = temp;
				else
					tail = temp;
				delete current;
				current = temp;	
				index--;
			}
			size--;
		}
	};
	 
	
	//************************
	//*** SEARCH FUNCTIONS ***
	//************************
	int close_search(void* search_data, int (*compare)(void*,void*)){		
		int value;
		
		value = (*compare)(search_data,get_data());
		if(value<0){					//Come in from above
			while( ( value = (*compare)( search_data,get_data() ) ) < 0){
				if(!prev())
					return -1;
			}
			if(value){
				next();
				return -1;
			}
			else
				return 1;
		}
		else if(value>0){				//Come in from below
			while( ( value = (*compare)( search_data,get_data() ) ) > 0){
				if(!next())
					return 0;
			}
			if(value){
				return -1;
			}
			else
				return 1;
		}
		else						//Exact Value
			return 1;
	};
	
	//Returns 1 on an exact match
	//Returns -1 on an in between match
	//Returns 0 if data is empty or if it is greater than the last item.
	int find_data(void* search_data, int (*compare)(void*,void*),float (*range)(void*,void*,void*)){
		long	new_index;
		float	hash;
  		long	from_begin;
    	long	from_end;
     	long	from_current;	
			
		if(!is_empty()){
			if(size > CLOSE){
				hash = (*range)(head->info,tail->info,search_data);
				new_index = (long) ( size * hash);
				from_current = labs( new_index - index);
				from_end = labs(new_index - size);
				from_begin = labs(new_index);
				if(from_begin<from_current)
					home();
				else if(from_end<from_current)
					end();
			}
			return close_search(search_data,(compare));
		}
		else
			return 0;
	};
	
	
	//***********************
	//*** QUERY FUNCTIONS ***
	//***********************
	int is_empty(){
 		if(head != 0)
			return 0;
		return 1;
  	};
  	
	int	is_home(){
 		if(!is_empty())
			if(current == head)
				return 1;
		return 0;
  	};
  	
	int is_end(){
 		if(!is_empty())
			if(current == tail)
				return 1;
		return 0;
  	};
  	
	long cur_index(){
		return index;
	};
	
	long get_size(){
		return size;
	};
	
	
	//***********************************
	//*** MARKING/DEMARKING FUNCTIONS ***
	//***********************************
	void* set_mark(){
		if(!is_empty())
			return ((void*) current);
	};

	//Sets current to the address pointed to by set_mark.
	void goto_mark(void* mark){
		current = (data*) mark;
		reindex();
	};
	
	void reindex(){
		data*	hold = current;
		
		home();
		while(hold != current)
			next();
		
	};

};

#endif
